﻿// MainViewModel.fs
// Copyright 2016, OSTRA / Takeshi Yoneki

namespace ViewModels

open System
open System.Windows
open System.Windows.Input
open System.Windows.Controls
open System.ComponentModel
open System.Collections.ObjectModel
open System.Runtime.CompilerServices

// XAML（MainWindow.xaml）に記述されたコントロールのBindingを解決するためのビューモデル。
// ビューモデルはXAMLで下記の記述によりインスタンスが作られWindowのDataContextに収められる。
// <Window.DataContext>
//   <local:MainViewModel/>
// </Window.DataContext>
type MainViewModel() =
    let mutable musicListInstance: MpdDB.MusicItem list = []
    let mutable musicCount = 0
    let MakeCounterText c a =
        c.ToString() + "/" + a.ToString()
    let mutable searchTextInstance = ""
    let mutable counterTextInstance = MakeCounterText musicCount musicCount

    // メニューコマンド
    let CreateMenuCommand name shortcut =
        if shortcut <> null then
            let cmdIG = new InputGestureCollection()
            let _ = cmdIG.Add(shortcut)
            new RoutedCommand(name, typeof<Window>, cmdIG)
        else
            new RoutedCommand(name, typeof<Window>)

    // イベントの扱い
    // http://stackoverflow.com/questions/1698147/f-using-inotifypropertychanged-for-data-binding
    // https://msdn.microsoft.com/visualfsharpdocs/conceptual/events-%5bfsharp%5d#
    // プロパティからコントロール方向のイベント通知に使う。
    // イベント通知はEvent.Triggerで行う。
    // 以下の記述はC#にてINotifyPropertyChangedを継承したクラスにおける
    // public event PropertyChangedEventHandler PropertyChanged;
    // に相当する。
    // キーワードとしてのeventが無いので、Eventというクラスを使うが、eventがそもそもおまじない感が強いところにF#専用のEventとか更なるおまじない。
    // そもそもeventはdelegateの一種だそうで、delegateはSystem.Delegateというクラスをキーワード化したもの。F#のEventはSystem.Delegateを使って実装してるんだろうと予想。
    // .NETの根幹を成す概念なんだからキーワードにしても良いのに、やはりGUI向けだから後回しにされた？
    // https://msdn.microsoft.com/en-us/library/edzehd2t(v=vs.110).aspx
    // CLIEventについては
    // https://msdn.microsoft.com/en-us/visualfsharpdocs/conceptual/core.clieventattribute-class-%5Bfsharp%5D
    let propertyChangedEvent = new Event<PropertyChangedEventHandler, PropertyChangedEventArgs>()
    interface INotifyPropertyChanged with
        [<CLIEvent>]
        member this.PropertyChanged = propertyChangedEvent.Publish

    // CallerMemberNameは機能しない。
    //member this.NotifyPropertyChanged ([<CallerMemberNameAttribute>] ?propertyName) =
    //    match propertyName with 
    //    | Some(name) ->
    //        propertyChangedEvent.Trigger(this, PropertyChangedEventArgs(name))
    //    | _->
    //        ()

    // XAMLでのCommand="{Binding PlayCommand}"の記述でMenuItemのCommandにインスタンスを用意するための記述。
    member val PlayCommand = CreateMenuCommand "PlayCommand" (KeyGesture(Key.Enter)) with get

    // XAMLでのCommand="{Binding ReloadCommand}"
    member val ReloadCommand = CreateMenuCommand "ReloadCommand" null with get

    // XAMLでのItemsSource="{Binding HitList}"
    // ObservableCollectionはメンバの変更があると通知をしてくれるらしい。
    member val Hits = new ObservableCollection<MpdDB.MusicItem>() with get

    // RecordListを設定あるいは置き換える。
    member this.SetRecordList(value) =
        musicListInstance <- value
        musicCount <- List.length musicListInstance
        this.ApplyHit()

    // 検索テキストによりMusicItemのリストをフィルタリング。
    member this.ApplyHit() =
        this.Hits.Clear()
        List.iter
            (fun (r: MpdDB.MusicItem) ->
                if (searchTextInstance.Length = 0 ||
                    r.Artist.IndexOf(this.SearchText, StringComparison.OrdinalIgnoreCase) <> -1 ||
                    r.Album.IndexOf(this.SearchText, StringComparison.OrdinalIgnoreCase) <> -1 ||
                    r.Title.IndexOf(this.SearchText, StringComparison.OrdinalIgnoreCase) <> -1) then
                        this.Hits.Add(r)
                    
            )
            musicListInstance
        this.CounterText <- MakeCounterText this.Hits.Count musicCount


    // XAMLでの<Binding Path="SearchText" UpdateSourceTrigger="PropertyChanged" />
    member this.SearchText
        with get () = searchTextInstance
        and set (value) = 
            // Text="{Binding SearchText, UpdateSourceTrigger=PropertyChanged}" により入力があったらここに来る。
            // このプロパティ（SearchText）を含むクラス（MainViewModel）がINotifyPropertyChangedの実装である必要はない。
            // コントロールからプロパティへ方向のイベント通知はXAMLの記述でOK。
            // コントロールのTextChangedイベントに直接ハンドラを置いた場合、漢字変換前にもイベントが発生するが、
            // PropertyChangedの方法だと漢字変換後に来る。
            // 直接ハンドラを置く方法で漢字変換後にするイベントがどれなのかあるいは存在するかは不明。
            searchTextInstance <- value
            this.ApplyHit()

    // XAMLでのText="{Binding CounterText}"
    member this.CounterText
        with get () = counterTextInstance
        and set (value) = 
            counterTextInstance <- value
            // Text="{Binding CounterText}"のコントロールに値の変化を通知。
            // INotifyPropertyChanged.PropertyChangedイベントで要求されるPropertyChangedEventHandlerの第二引数がPropertyChangedEventArgs
            // PropertyChangedEventArgsの文字列"CounterText"で{Binding CounterText}を探すようだ。
            propertyChangedEvent.Trigger(this, new PropertyChangedEventArgs("CounterText"))
            //this.NotifyPropertyChanged()
